import os


ROOT_DIR = os.path.abspath(os.sep)
ERROR_LOG = "sm_tools_error.log"

# These numbers in a stepchart indicate notes that require stepping on
# 1 = note, 2 = hold, 4 = roll
NOTE_TYPES = ["1", "2", "4"]
